@@ -120,6 +120,12 @@ impl TryCastFrom<Value> for AxisBounds {
     }
 }
 
+impl fmt::Debug for AxisBounds {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        fmt::Display::fmt(self, f)
+    }
+}
+
 impl fmt::Display for AxisBounds {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         use AxisBounds::*;
@@ -210,8 +216,7 @@ impl Bounds {
     /// ```
     /// # use tc_tensor::{Bounds, Shape};
     /// let mut bounds = Bounds::from(&[0u64][..]);
-    /// bounds.normalize(&Shape::from(vec![2, 3, 4]));
-    /// assert_eq!(bounds.to_shape(), Shape::from(vec![3, 4]));
+    /// assert_eq!(bounds.to_shape(&Shape::from(vec![2, 3, 4])).unwrap(), Shape::from(vec![3, 4]));
     /// ```
     pub fn normalize(&mut self, shape: &Shape) {
         assert!(self.len() <= shape.len());
@@ -339,6 +344,20 @@ impl TryCastFrom<Value> for Bounds {
     }
 }
 
+impl fmt::Debug for Bounds {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        write!(
+            f,
+            "[{}]",
+            self.axes
+                .iter()
+                .map(|axis| format!("{:?}", axis))
+                .collect::<Vec<String>>()
+                .join(", ")
+        )
+    }
+}
+
 impl fmt::Display for Bounds {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         write!(
@@ -420,6 +439,17 @@ impl Shape {
         self.0.iter().product()
     }
 
+    /// Return a `TCError` if any of the given axes is out of bounds.
+    pub fn validate_axes(&self, axes: &[usize]) -> TCResult<()> {
+        match axes.iter().max() {
+            Some(max) if max > &self.len() => Err(TCError::unsupported(format!(
+                "shape {} has no axis {}",
+                self, max
+            ))),
+            _ => Ok(()),
+        }
+    }
+
     /// Return a `TCError` if the given `Bounds` don't fit within this `Shape`.
     pub fn validate_bounds(&self, bounds: &Bounds) -> TCResult<()> {
         if self.contains_bounds(bounds) {
