@@ -1,6 +1,6 @@
 use std::convert::TryFrom;
 
-use afarray::{Array, ArrayExt, Coords};
+use afarray::{Array, ArrayExt, Coords, Offsets};
 use async_trait::async_trait;
 use futures::future::{self, TryFutureExt};
 use futures::stream::{self, StreamExt, TryStreamExt};
@@ -12,21 +12,17 @@ use tc_error::*;
 use tc_transact::fs::{Dir, File};
 use tc_transact::{Transaction, TxnId};
 use tc_value::{Number, NumberClass, NumberInstance, NumberType};
-use tcgeneric::{TCBoxTryFuture, TCStream, TCTryStream};
+use tcgeneric::{TCBoxStream, TCBoxTryFuture, TCBoxTryStream};
 
 use crate::sparse::{SparseAccess, SparseAccessor};
 use crate::stream::{Read, ReadValueAt};
 use crate::transform;
-use crate::{Bounds, Coord, Phantom, Shape, TensorAccess, TensorType, ERR_NONBIJECTIVE_WRITE};
+use crate::{Bounds, Coord, Phantom, Shape, TensorAccess, TensorType};
 
 use super::file::{BlockListFile, BlockListFileSlice};
 use super::stream::SparseValueStream;
 use super::{DenseTensor, PER_BLOCK};
 
-const ERR_NON_SEQUENTIAL_WRITE: &str =
-    "non-sequential writes to a dense Tensor are not efficient--\
-consider transposing or broadcasting the value instead";
-
 /// Common [`DenseTensor`] access methods
 #[async_trait]
 pub trait DenseAccess<FD: File<Array>, FS: File<Node>, D: Dir, T: Transaction<D>>:
@@ -42,7 +38,7 @@ pub trait DenseAccess<FD: File<Array>, FS: File<Node>, D: Dir, T: Transaction<D>
     fn accessor(self) -> DenseAccessor<FD, FS, D, T>;
 
     /// Return a stream of the [`Array`]s which this [`DenseTensor`] comprises.
-    fn block_stream<'a>(self, txn: Self::Txn) -> TCBoxTryFuture<'a, TCTryStream<'a, Array>> {
+    fn block_stream<'a>(self, txn: Self::Txn) -> TCBoxTryFuture<'a, TCBoxTryStream<'a, Array>> {
         Box::pin(async move {
             let blocks = self.value_stream(txn).await?;
             let blocks = blocks
@@ -50,13 +46,13 @@ pub trait DenseAccess<FD: File<Array>, FS: File<Node>, D: Dir, T: Transaction<D>
                 .map(|values| values.into_iter().collect::<TCResult<Vec<Number>>>())
                 .map_ok(Array::from);
 
-            let blocks: TCTryStream<'a, Array> = Box::pin(blocks);
+            let blocks: TCBoxTryStream<'a, Array> = Box::pin(blocks);
             Ok(blocks)
         })
     }
 
     /// Return a stream of the elements of this [`DenseTensor`].
-    fn value_stream<'a>(self, txn: Self::Txn) -> TCBoxTryFuture<'a, TCTryStream<'a, Number>> {
+    fn value_stream<'a>(self, txn: Self::Txn) -> TCBoxTryFuture<'a, TCBoxTryStream<'a, Number>> {
         Box::pin(async move {
             let values = self.block_stream(txn).await?;
 
@@ -66,7 +62,7 @@ pub trait DenseAccess<FD: File<Array>, FS: File<Node>, D: Dir, T: Transaction<D>
                 .map_ok(futures::stream::iter)
                 .try_flatten();
 
-            let values: TCTryStream<'a, Number> = Box::pin(values);
+            let values: TCBoxTryStream<'a, Number> = Box::pin(values);
             Ok(values)
         })
     }
@@ -79,9 +75,20 @@ pub trait DenseAccess<FD: File<Array>, FS: File<Node>, D: Dir, T: Transaction<D>
 
     /// Return an Array with the values at the given coordinates.
     async fn read_values(self, txn: Self::Txn, coords: Coords) -> TCResult<Array>;
+}
 
+/// Common [`DenseTensor`] access methods
+#[async_trait]
+pub trait DenseWrite<FD: File<Array>, FS: File<Node>, D: Dir, T: Transaction<D>>:
+    DenseAccess<FD, FS, D, T>
+{
     /// Overwrite this accessor's contents with those of the given accessor.
-    async fn write<V: DenseAccess<FD, FS, D, T>>(&self, txn: Self::Txn, value: V) -> TCResult<()>;
+    async fn write<V: DenseAccess<FD, FS, D, T>>(
+        &self,
+        txn: Self::Txn,
+        bounds: Bounds,
+        value: V,
+    ) -> TCResult<()>;
 
     /// Write a value to the slice of this [`DenseTensor`] with the given [`Bounds`].
     async fn write_value(&self, txn_id: TxnId, bounds: Bounds, number: Number) -> TCResult<()>;
@@ -187,7 +194,7 @@ where
         self
     }
 
-    fn block_stream<'a>(self, txn: T) -> TCBoxTryFuture<'a, TCTryStream<'a, Array>> {
+    fn block_stream<'a>(self, txn: T) -> TCBoxTryFuture<'a, TCBoxTryStream<'a, Array>> {
         match self {
             Self::File(file) => file.block_stream(txn),
             Self::Slice(slice) => slice.block_stream(txn),
@@ -202,7 +209,7 @@ where
         }
     }
 
-    fn value_stream<'a>(self, txn: T) -> TCBoxTryFuture<'a, TCTryStream<'a, Number>> {
+    fn value_stream<'a>(self, txn: T) -> TCBoxTryFuture<'a, TCBoxTryStream<'a, Number>> {
         match self {
             Self::File(file) => file.value_stream(txn),
             Self::Slice(slice) => slice.value_stream(txn),
@@ -281,34 +288,33 @@ where
             Self::Unary(unary) => unary.read_values(txn, coords).await,
         }
     }
+}
 
-    async fn write<V: DenseAccess<FD, FS, D, T>>(&self, txn: Self::Txn, value: V) -> TCResult<()> {
+#[async_trait]
+impl<FD, FS, D, T> DenseWrite<FD, FS, D, T> for DenseAccessor<FD, FS, D, T>
+where
+    FD: File<Array> + TryFrom<D::File, Error = TCError>,
+    FS: File<Node> + TryFrom<D::File, Error = TCError>,
+    D: Dir,
+    T: Transaction<D>,
+    D::FileClass: From<TensorType>,
+{
+    async fn write<V: DenseAccess<FD, FS, D, T>>(
+        &self,
+        txn: Self::Txn,
+        bounds: Bounds,
+        value: V,
+    ) -> TCResult<()> {
         match self {
-            Self::File(file) => file.write(txn, value).await,
-            Self::Slice(slice) => slice.write(txn, value).await,
-            Self::Broadcast(broadcast) => broadcast.write(txn, value).await,
-            Self::Cast(cast) => cast.write(txn, value).await,
-            Self::Combine(combine) => combine.write(txn, value).await,
-            Self::Expand(expansion) => expansion.write(txn, value).await,
-            Self::Reduce(reduced) => reduced.write(txn, value).await,
-            Self::Sparse(sparse) => sparse.write(txn, value).await,
-            Self::Transpose(transpose) => transpose.write(txn, value).await,
-            Self::Unary(unary) => unary.write(txn, value).await,
+            Self::File(file) => file.write(txn, bounds, value).await,
+            _ => Err(TCError::unsupported("cannot write to a Tensor view")),
         }
     }
 
     async fn write_value(&self, txn_id: TxnId, bounds: Bounds, number: Number) -> TCResult<()> {
         match self {
             Self::File(file) => file.write_value(txn_id, bounds, number).await,
-            Self::Slice(slice) => slice.write_value(txn_id, bounds, number).await,
-            Self::Broadcast(broadcast) => broadcast.write_value(txn_id, bounds, number).await,
-            Self::Cast(cast) => cast.write_value(txn_id, bounds, number).await,
-            Self::Combine(combine) => combine.write_value(txn_id, bounds, number).await,
-            Self::Expand(expansion) => expansion.write_value(txn_id, bounds, number).await,
-            Self::Reduce(reduced) => reduced.write_value(txn_id, bounds, number).await,
-            Self::Sparse(sparse) => sparse.write_value(txn_id, bounds, number).await,
-            Self::Transpose(transpose) => transpose.write_value(txn_id, bounds, number).await,
-            Self::Unary(unary) => unary.write_value(txn_id, bounds, number).await,
+            _ => Err(TCError::unsupported("cannot write to a Tensor view")),
         }
     }
 }
@@ -450,7 +456,7 @@ where
         DenseAccessor::Combine(Box::new(combine))
     }
 
-    fn block_stream<'a>(self, txn: T) -> TCBoxTryFuture<'a, TCTryStream<'a, Array>> {
+    fn block_stream<'a>(self, txn: T) -> TCBoxTryFuture<'a, TCBoxTryStream<'a, Array>> {
         debug!("BlockListCombine::block_stream");
 
         Box::pin(async move {
@@ -464,7 +470,7 @@ where
                 .map(|(l, r)| Ok((l?, r?)))
                 .map_ok(move |(l, r)| combinator(&l, &r));
 
-            let blocks: TCTryStream<'a, Array> = Box::pin(blocks);
+            let blocks: TCBoxTryStream<'a, Array> = Box::pin(blocks);
             Ok(blocks)
         })
     }
@@ -483,6 +489,13 @@ where
     }
 
     fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self::Transpose> {
+        debug!(
+            "BlockListCombine::transpose {} {} {:?}",
+            self.left.shape(),
+            self.right.shape(),
+            permutation
+        );
+
         let left = self.left.transpose(permutation.clone())?;
         let right = self.right.transpose(permutation)?;
 
@@ -503,18 +516,6 @@ where
 
         Ok((self.combinator)(&left, &right))
     }
-
-    async fn write<V: DenseAccess<FD, FS, D, T>>(
-        &self,
-        _txn: Self::Txn,
-        _value: V,
-    ) -> TCResult<()> {
-        Err(TCError::unsupported(ERR_NONBIJECTIVE_WRITE))
-    }
-
-    async fn write_value(&self, _txn_id: TxnId, _bounds: Bounds, _number: Number) -> TCResult<()> {
-        Err(TCError::unsupported(ERR_NONBIJECTIVE_WRITE))
-    }
 }
 
 impl<FD, FS, D, T, L, R> ReadValueAt<D> for BlockListCombine<FD, FS, D, T, L, R>
@@ -578,7 +579,7 @@ where
     }
 
     fn ndim(&self) -> usize {
-        self.source.ndim()
+        self.shape().len()
     }
 
     fn shape(&'_ self) -> &'_ Shape {
@@ -614,7 +615,7 @@ where
         DenseAccessor::Broadcast(Box::new(broadcast))
     }
 
-    fn block_stream<'a>(self, txn: T) -> TCBoxTryFuture<'a, TCTryStream<'a, Array>> {
+    fn block_stream<'a>(self, txn: T) -> TCBoxTryFuture<'a, TCBoxTryStream<'a, Array>> {
         let shape = self.shape().clone();
         let size = self.size();
         let rebase = self.rebase;
@@ -634,7 +635,7 @@ where
             .map(move |coords| source.clone().read_values(txn.clone(), coords))
             .buffered(num_cpus::get());
 
-        let blocks: TCTryStream<'a, Array> = Box::pin(blocks);
+        let blocks: TCBoxTryStream<'a, Array> = Box::pin(blocks);
         Box::pin(future::ready(Ok(blocks)))
     }
 
@@ -648,6 +649,12 @@ where
     }
 
     fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self::Transpose> {
+        debug!(
+            "BlockListTranspose {:?} (shape is {})",
+            permutation,
+            self.shape()
+        );
+
         BlockListTranspose::new(self, permutation)
     }
 
@@ -655,18 +662,6 @@ where
         let coords = self.rebase.invert_coords(&coords);
         self.source.read_values(txn, coords).await
     }
-
-    async fn write<V: DenseAccess<FD, FS, D, T>>(
-        &self,
-        _txn: Self::Txn,
-        _value: V,
-    ) -> TCResult<()> {
-        Err(TCError::unsupported(ERR_NONBIJECTIVE_WRITE))
-    }
-
-    async fn write_value(&self, _txn_id: TxnId, _bounds: Bounds, _number: Number) -> TCResult<()> {
-        Err(TCError::unsupported(ERR_NONBIJECTIVE_WRITE))
-    }
 }
 
 impl<FD, FS, D, T, B> ReadValueAt<D> for BlockListBroadcast<FD, FS, D, T, B>
@@ -758,12 +753,12 @@ where
         DenseAccessor::Cast(Box::new(cast))
     }
 
-    fn block_stream<'a>(self, txn: T) -> TCBoxTryFuture<'a, TCTryStream<'a, Array>> {
+    fn block_stream<'a>(self, txn: T) -> TCBoxTryFuture<'a, TCBoxTryStream<'a, Array>> {
         Box::pin(async move {
             let dtype = self.dtype;
-            let blocks: TCStream<'a, TCResult<Array>> = self.source.block_stream(txn).await?;
+            let blocks: TCBoxStream<'a, TCResult<Array>> = self.source.block_stream(txn).await?;
             let cast = blocks.map_ok(move |array| array.cast_into(dtype));
-            let cast: TCTryStream<'a, Array> = Box::pin(cast);
+            let cast: TCBoxTryStream<'a, Array> = Box::pin(cast);
             Ok(cast)
         })
     }
@@ -786,14 +781,6 @@ where
             .map_ok(|values| values.cast_into(dtype))
             .await
     }
-
-    async fn write<V: DenseAccess<FD, FS, D, T>>(&self, txn: Self::Txn, value: V) -> TCResult<()> {
-        self.source.write(txn, value).await
-    }
-
-    async fn write_value(&self, txn_id: TxnId, bounds: Bounds, number: Number) -> TCResult<()> {
-        self.source.write_value(txn_id, bounds, number).await
-    }
 }
 
 impl<FD, FS, D, T, B> ReadValueAt<D> for BlockListCast<FD, FS, D, T, B>
@@ -891,15 +878,17 @@ where
         DenseAccessor::Expand(Box::new(expand))
     }
 
-    fn block_stream<'a>(self, txn: T) -> TCBoxTryFuture<'a, TCTryStream<'a, Array>> {
+    fn block_stream<'a>(self, txn: T) -> TCBoxTryFuture<'a, TCBoxTryStream<'a, Array>> {
         self.source.block_stream(txn)
     }
 
-    fn value_stream<'a>(self, txn: T) -> TCBoxTryFuture<'a, TCTryStream<'a, Number>> {
+    fn value_stream<'a>(self, txn: T) -> TCBoxTryFuture<'a, TCBoxTryStream<'a, Number>> {
         self.source.value_stream(txn)
     }
 
-    fn slice(self, bounds: Bounds) -> TCResult<Self::Slice> {
+    fn slice(self, mut bounds: Bounds) -> TCResult<Self::Slice> {
+        self.shape().validate_bounds(&bounds)?;
+        bounds.normalize(self.shape());
         let bounds = self.rebase.invert_bounds(bounds);
         self.source.slice(bounds) // TODO: expand the result
     }
@@ -913,15 +902,6 @@ where
         let coords = self.rebase.invert_coords(&coords);
         self.source.read_values(txn, coords).await
     }
-
-    async fn write<V: DenseAccess<FD, FS, D, T>>(&self, txn: Self::Txn, value: V) -> TCResult<()> {
-        self.source.write(txn, value).await
-    }
-
-    async fn write_value(&self, txn_id: TxnId, bounds: Bounds, number: Number) -> TCResult<()> {
-        let bounds = self.rebase.invert_bounds(bounds);
-        self.source.write_value(txn_id, bounds, number).await
-    }
 }
 
 impl<FD, FS, D, T, B> ReadValueAt<D> for BlockListExpand<FD, FS, D, T, B>
@@ -1012,7 +992,7 @@ where
     D::FileClass: From<TensorType>,
 {
     type Slice = BlockListReduce<FD, FS, D, T, <B as DenseAccess<FD, FS, D, T>>::Slice>;
-    type Transpose = BlockListReduce<FD, FS, D, T, <B as DenseAccess<FD, FS, D, T>>::Transpose>;
+    type Transpose = BlockListTranspose<FD, FS, D, T, Self>;
 
     fn accessor(self) -> DenseAccessor<FD, FS, D, T> {
         let reduce = BlockListReduce {
@@ -1024,7 +1004,7 @@ where
         DenseAccessor::Reduce(Box::new(reduce))
     }
 
-    fn value_stream<'a>(self, txn: T) -> TCBoxTryFuture<'a, TCTryStream<'a, Number>> {
+    fn value_stream<'a>(self, txn: T) -> TCBoxTryFuture<'a, TCBoxTryStream<'a, Number>> {
         Box::pin(async move {
             let values = stream::iter(Bounds::all(self.shape()).affected())
                 .map(move |coord| {
@@ -1039,12 +1019,13 @@ where
                 })
                 .buffered(num_cpus::get());
 
-            let values: TCTryStream<'a, Number> = Box::pin(values);
+            let values: TCBoxTryStream<'a, Number> = Box::pin(values);
             Ok(values)
         })
     }
 
     fn slice(self, bounds: Bounds) -> TCResult<Self::Slice> {
+        self.shape().validate_bounds(&bounds)?;
         let reduce_axis = self.rebase.reduce_axis(&bounds);
         let source_bounds = self.rebase.invert_bounds(bounds);
         let slice = self.source.slice(source_bounds)?;
@@ -1052,11 +1033,13 @@ where
     }
 
     fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self::Transpose> {
-        let permutation = permutation.unwrap_or_else(|| (0..self.ndim()).rev().collect());
-        let reduce_axis = self.rebase.reduce_axis(&Bounds::all(self.shape()));
-        let source_axes = self.rebase.invert_axes(permutation);
-        let source = self.source.transpose(Some(source_axes))?;
-        BlockListReduce::new(source, reduce_axis, self.reductor)
+        debug!(
+            "BlockListReduce::transpose {} {:?}",
+            self.shape(),
+            permutation
+        );
+
+        BlockListTranspose::new(self, permutation)
     }
 
     async fn read_values(self, txn: Self::Txn, coords: Coords) -> TCResult<Array> {
@@ -1070,18 +1053,6 @@ where
 
         Ok(Array::from(values))
     }
-
-    async fn write<V: DenseAccess<FD, FS, D, T>>(
-        &self,
-        _txn: Self::Txn,
-        _value: V,
-    ) -> TCResult<()> {
-        Err(TCError::unsupported(ERR_NONBIJECTIVE_WRITE))
-    }
-
-    async fn write_value(&self, _txn_id: TxnId, _bounds: Bounds, _number: Number) -> TCResult<()> {
-        Err(TCError::unsupported(ERR_NONBIJECTIVE_WRITE))
-    }
 }
 
 impl<FD, FS, D, T, B> ReadValueAt<D> for BlockListReduce<FD, FS, D, T, B>
@@ -1097,6 +1068,7 @@ where
 
     fn read_value_at<'a>(self, txn: Self::Txn, coord: Coord) -> Read<'a> {
         Box::pin(async move {
+            self.shape().validate_coord(&coord)?;
             let reductor = self.reductor;
             let source_bounds = self.rebase.invert_coord(&coord);
             let slice = self.source.slice(source_bounds)?;
@@ -1167,8 +1139,8 @@ where
     B: DenseAccess<FD, FS, D, T>,
     D::FileClass: From<TensorType>,
 {
-    type Slice = BlockListTranspose<FD, FS, D, T, B::Slice>;
-    type Transpose = Self;
+    type Slice = <<B as DenseAccess<FD, FS, D, T>>::Slice as DenseAccess<FD, FS, D, T>>::Transpose;
+    type Transpose = B::Transpose;
 
     fn accessor(self) -> DenseAccessor<FD, FS, D, T> {
         let accessor = BlockListTranspose {
@@ -1180,35 +1152,54 @@ where
         DenseAccessor::Transpose(Box::new(accessor))
     }
 
-    fn block_stream<'a>(self, _txn: T) -> TCBoxTryFuture<'a, TCTryStream<'a, Array>> {
-        Box::pin(future::ready(Err(TCError::not_implemented(
-            "BlockListTranspose::block_stream",
-        ))))
+    fn block_stream<'a>(self, txn: T) -> TCBoxTryFuture<'a, TCBoxTryStream<'a, Array>> {
+        Box::pin(async move {
+            let size = self.size();
+            let shape = self.shape().clone();
+
+            let per_block = PER_BLOCK as u64;
+            let blocks = stream::iter((0..size).step_by(PER_BLOCK))
+                .map(move |start| {
+                    let end = start + per_block;
+                    if end > size {
+                        (start, size)
+                    } else {
+                        (start, end)
+                    }
+                })
+                .map(|(start, end)| Offsets::range(start, end))
+                .map(move |offsets| Coords::from_offsets(offsets, &shape))
+                .map(move |coords| self.clone().read_values(txn.clone(), coords))
+                .buffered(num_cpus::get());
+
+            let blocks: TCBoxTryStream<'a, Array> = Box::pin(blocks);
+            Ok(blocks)
+        })
     }
 
-    fn slice(self, _bounds: Bounds) -> TCResult<Self::Slice> {
-        Err(TCError::not_implemented("BlockListTranspose::slice"))
+    fn slice(self, mut bounds: Bounds) -> TCResult<Self::Slice> {
+        bounds.normalize(self.shape());
+        let permutation = self.rebase.invert_permutation(&bounds);
+        let source_bounds = self.rebase.invert_bounds(&bounds);
+        let expected_shape = source_bounds.to_shape(self.source.shape())?;
+        let slice = self.source.slice(source_bounds)?;
+        debug_assert_eq!(slice.shape(), &expected_shape);
+        slice.transpose(Some(permutation))
     }
 
-    fn transpose(self, _permutation: Option<Vec<usize>>) -> TCResult<Self::Transpose> {
-        Err(TCError::not_implemented("BlockListTranspose::transpose"))
+    fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self::Transpose> {
+        let permutation = if let Some(permutation) = permutation {
+            self.rebase.invert_axes(permutation)
+        } else {
+            self.rebase.invert_axes((0..self.ndim()).rev().collect())
+        };
+
+        self.source.transpose(Some(permutation))
     }
 
-    async fn read_values(self, _txn: Self::Txn, _coords: Coords) -> TCResult<Array> {
-        Err(TCError::not_implemented("BlockListTranspose::read_values"))
-    }
-
-    async fn write<V: DenseAccess<FD, FS, D, T>>(
-        &self,
-        _txn: Self::Txn,
-        _value: V,
-    ) -> TCResult<()> {
-        Err(TCError::unsupported(ERR_NON_SEQUENTIAL_WRITE))
-    }
-
-    async fn write_value(&self, txn_id: TxnId, bounds: Bounds, number: Number) -> TCResult<()> {
-        let bounds = self.rebase.invert_bounds(bounds);
-        self.source.write_value(txn_id, bounds, number).await
+    async fn read_values(self, txn: Self::Txn, coords: Coords) -> TCResult<Array> {
+        let coords = self.rebase.invert_coords(&coords);
+        self.source.read_values(txn, coords).await
     }
 }
 
@@ -1224,13 +1215,14 @@ where
     type Txn = T;
 
     fn read_value_at<'a>(self, txn: Self::Txn, coord: Coord) -> Read<'a> {
-        let source_coord = self.rebase.invert_coord(&coord);
-        let read = self
-            .source
-            .read_value_at(txn, source_coord)
-            .map_ok(|(_, val)| (coord, val));
-
-        Box::pin(read)
+        Box::pin(async move {
+            self.shape().validate_coord(&coord)?;
+            let source_coord = self.rebase.invert_coord(&coord);
+            self.source
+                .read_value_at(txn, source_coord)
+                .map_ok(|(_, val)| (coord, val))
+                .await
+        })
     }
 }
 
@@ -1286,7 +1278,7 @@ where
         })
     }
 
-    fn value_stream<'a>(self, txn: T) -> TCBoxTryFuture<'a, TCTryStream<'a, Number>> {
+    fn value_stream<'a>(self, txn: T) -> TCBoxTryFuture<'a, TCBoxTryStream<'a, Number>> {
         debug!("BlockListSparse::value_stream");
 
         Box::pin(async move {
@@ -1294,12 +1286,13 @@ where
             let zero = self.dtype().zero();
             let filled = self.source.filled(txn).await?;
             let values = SparseValueStream::new(filled, bounds, zero).await?;
-            let values: TCTryStream<'a, Number> = Box::pin(values);
+            let values: TCBoxTryStream<'a, Number> = Box::pin(values);
             Ok(values)
         })
     }
 
     fn slice(self, bounds: Bounds) -> TCResult<Self::Slice> {
+        self.shape().validate_bounds(&bounds)?;
         let slice = self.source.slice(bounds)?;
         Ok(slice.into())
     }
@@ -1321,36 +1314,6 @@ where
 
         Ok(Array::from(values))
     }
-
-    async fn write<V: DenseAccess<FD, FS, D, T>>(&self, txn: Self::Txn, value: V) -> TCResult<()> {
-        if value.shape() != self.shape() {
-            return Err(TCError::unsupported(format!(
-                "cannot write a value of shape {} to a tensor of shape {}",
-                value.shape(),
-                self.shape()
-            )));
-        }
-
-        let txn_id = *txn.id();
-        let bounds = Bounds::all(self.shape());
-        let values = value.value_stream(txn).await?;
-        let source = &self.source;
-        stream::iter(bounds.affected())
-            .zip(values)
-            .map(|(coord, r)| r.map(|n| (coord, n)))
-            .map_ok(|(coord, n)| source.write_value(txn_id, coord, n))
-            .try_buffer_unordered(num_cpus::get())
-            .try_fold((), |_, _| future::ready(Ok(())))
-            .await
-    }
-
-    async fn write_value(&self, txn_id: TxnId, bounds: Bounds, number: Number) -> TCResult<()> {
-        stream::iter(bounds.affected())
-            .map(|coord| self.source.write_value(txn_id, coord, number))
-            .buffer_unordered(num_cpus::get())
-            .try_fold((), |_, _| future::ready(Ok(())))
-            .await
-    }
 }
 
 impl<FD, FS, D, T, A> ReadValueAt<D> for BlockListSparse<FD, FS, D, T, A>
@@ -1460,11 +1423,11 @@ where
         DenseAccessor::Unary(Box::new(unary))
     }
 
-    fn block_stream<'a>(self, txn: T) -> TCBoxTryFuture<'a, TCTryStream<'a, Array>> {
+    fn block_stream<'a>(self, txn: T) -> TCBoxTryFuture<'a, TCBoxTryStream<'a, Array>> {
         Box::pin(async move {
             let transform = self.transform;
             let blocks = self.source.block_stream(txn).await?;
-            let blocks: TCTryStream<'a, Array> =
+            let blocks: TCBoxTryStream<'a, Array> =
                 Box::pin(blocks.map_ok(move |array| transform(&array)));
 
             Ok(blocks)
@@ -1501,18 +1464,6 @@ where
             .map_ok(move |values| (transform)(&values))
             .await
     }
-
-    async fn write<V: DenseAccess<FD, FS, D, T>>(
-        &self,
-        _txn: Self::Txn,
-        _value: V,
-    ) -> TCResult<()> {
-        Err(TCError::unsupported(ERR_NONBIJECTIVE_WRITE))
-    }
-
-    async fn write_value(&self, _txn_id: TxnId, _bounds: Bounds, _number: Number) -> TCResult<()> {
-        Err(TCError::unsupported(ERR_NONBIJECTIVE_WRITE))
-    }
 }
 
 impl<FD, FS, D, T, B> ReadValueAt<D> for BlockListUnary<FD, FS, D, T, B>
