@@ -16,7 +16,7 @@ use tc_error::*;
 use tc_transact::fs::{CopyFrom, Dir, File, Hash, Persist, Restore};
 use tc_transact::{IntoView, Transact, Transaction, TxnId};
 use tc_value::{Number, NumberClass, NumberInstance, NumberType};
-use tcgeneric::{TCBoxTryFuture, TCTryStream};
+use tcgeneric::{TCBoxTryFuture, TCBoxTryStream};
 
 use super::sparse::{DenseToSparse, SparseTensor};
 use super::stream::{Read, ReadValueAt};
@@ -27,7 +27,7 @@ use super::{
 };
 
 use access::*;
-pub use access::{BlockListSparse, DenseAccess, DenseAccessor};
+pub use access::{BlockListSparse, DenseAccess, DenseAccessor, DenseWrite};
 pub use file::BlockListFile;
 use futures::StreamExt;
 
@@ -351,7 +351,7 @@ where
     FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
-    B: DenseAccess<FD, FS, D, T>,
+    B: DenseWrite<FD, FS, D, T>,
     D::FileClass: From<TensorType>,
 {
     type Txn = T;
@@ -384,7 +384,7 @@ where
     FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
-    B: DenseAccess<FD, FS, D, T>,
+    B: DenseWrite<FD, FS, D, T>,
     O: DenseAccess<FD, FS, D, T>,
     D::FileClass: From<TensorType>,
 {
@@ -401,21 +401,7 @@ where
         other: DenseTensor<FD, FS, D, T, O>,
     ) -> TCResult<()> {
         debug!("write {} to dense {}", other, bounds);
-        if bounds == Bounds::all(self.shape()) {
-            self.blocks.write(txn, other.blocks).await
-        } else {
-            let slice = self.slice(bounds)?;
-            if other.shape() != slice.shape() {
-                return Err(TCError::unsupported(format!(
-                    "cannot write a tensor of shape {} to a slice of shape {}",
-                    other.shape(),
-                    slice.shape()
-                )));
-            }
-
-            let bounds = Bounds::all(slice.shape());
-            slice.write(txn, bounds, other).await
-        }
+        self.blocks.write(txn, bounds, other.blocks).await
     }
 }
 
@@ -426,7 +412,7 @@ where
     FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
-    B: DenseAccess<FD, FS, D, T>,
+    B: DenseWrite<FD, FS, D, T>,
     D::FileClass: From<TensorType>,
 {
     type Txn = T;
@@ -757,7 +743,7 @@ where
     type Item = Array;
     type Txn = T;
 
-    async fn hashable(&'en self, txn: &'en T) -> TCResult<TCTryStream<'en, Self::Item>> {
+    async fn hashable(&'en self, txn: &'en T) -> TCResult<TCBoxTryStream<'en, Self::Item>> {
         self.blocks.clone().block_stream(txn.clone()).await
     }
 }
@@ -773,7 +759,11 @@ where
     B: DenseAccess<FD, FS, D, T>,
     D::FileClass: From<TensorType>,
 {
-    async fn copy_from(instance: DenseTensor<FD, FS, D, T, B>, file: FD, txn: T) -> TCResult<Self> {
+    async fn copy_from(
+        instance: DenseTensor<FD, FS, D, T, B>,
+        file: FD,
+        txn: &T,
+    ) -> TCResult<Self> {
         BlockListFile::copy_from(instance.blocks, file, txn)
             .map_ok(Self::from)
             .await
@@ -957,7 +947,7 @@ impl<'en> en::IntoStream<'en> for DenseTensorView<'en> {
 
 struct BlockStreamView<'en> {
     dtype: NumberType,
-    blocks: TCTryStream<'en, Array>,
+    blocks: TCBoxTryStream<'en, Array>,
 }
 
 impl<'en> en::IntoStream<'en> for BlockStreamView<'en> {
@@ -967,7 +957,7 @@ impl<'en> en::IntoStream<'en> for BlockStreamView<'en> {
         };
 
         fn encodable<'en, T: af::HasAfEnum + Clone + Default + 'en>(
-            blocks: TCTryStream<'en, Array>,
+            blocks: TCBoxTryStream<'en, Array>,
         ) -> impl Stream<Item = Vec<T>> + 'en {
             // an error can't be encoded within an array
             // so in case of a read error, let the receiver figure out that the tensor
@@ -1006,7 +996,7 @@ impl<'en> en::IntoStream<'en> for BlockStreamView<'en> {
     }
 }
 
-fn encodable_c32<'en>(blocks: TCTryStream<'en, Array>) -> impl Stream<Item = Vec<f32>> + 'en {
+fn encodable_c32<'en>(blocks: TCBoxTryStream<'en, Array>) -> impl Stream<Item = Vec<f32>> + 'en {
     blocks
         .take_while(|r| future::ready(r.is_ok()))
         .map(|block| block.expect("tensor block"))
@@ -1027,7 +1017,7 @@ fn encodable_c32<'en>(blocks: TCTryStream<'en, Array>) -> impl Stream<Item = Vec
         })
 }
 
-fn encodable_c64<'en>(blocks: TCTryStream<'en, Array>) -> impl Stream<Item = Vec<f64>> + 'en {
+fn encodable_c64<'en>(blocks: TCBoxTryStream<'en, Array>) -> impl Stream<Item = Vec<f64>> + 'en {
     blocks
         .take_while(|r| future::ready(r.is_ok()))
         .map(|block| block.expect("tensor block"))
