@@ -12,14 +12,13 @@ use tc_error::*;
 use tc_transact::fs::{Dir, File};
 use tc_transact::{Transaction, TxnId};
 use tc_value::{Number, NumberClass, NumberInstance, NumberType};
-use tcgeneric::{TCBoxTryFuture, TCTryStream};
+use tcgeneric::{TCBoxTryFuture, TCBoxTryStream};
 
 use crate::dense::{DenseAccess, DenseAccessor, DenseTensor, PER_BLOCK};
 use crate::stream::{sorted_coords, sorted_values, Read, ReadValueAt};
 use crate::transform;
 use crate::{
     coord_bounds, AxisBounds, Bounds, Coord, Phantom, Shape, TensorAccess, TensorType, TensorUnary,
-    ERR_NONBIJECTIVE_WRITE,
 };
 
 use super::combine::{coord_to_offset, SparseCombine};
@@ -44,7 +43,9 @@ pub trait SparseAccess<FD: File<Array>, FS: File<Node>, D: Dir, T: Transaction<D
     async fn filled<'a>(self, txn: T) -> TCResult<SparseStream<'a>>;
 
     /// Return an ordered stream of unique [`Coord`]s on the given axes with nonzero values.
-    async fn filled_at<'a>(self, txn: T, axes: Vec<usize>) -> TCResult<TCTryStream<'a, Coords>>;
+    ///
+    /// TODO: add `sort` parameter to make sorting results optional.
+    async fn filled_at<'a>(self, txn: T, axes: Vec<usize>) -> TCResult<TCBoxTryStream<'a, Coords>>;
 
     /// Return the number of nonzero values in this [`SparseTensor`].
     async fn filled_count(self, txn: T) -> TCResult<u64>;
@@ -59,7 +60,13 @@ pub trait SparseAccess<FD: File<Array>, FS: File<Node>, D: Dir, T: Transaction<D
     ///
     /// If no permutation is given, this accessor's axes will be reversed.
     fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self::Transpose>;
+}
 
+/// Write methods for [`SparseTensor`] data
+#[async_trait]
+pub trait SparseWrite<FD: File<Array>, FS: File<Node>, D: Dir, T: Transaction<D>>:
+    SparseAccess<FD, FS, D, T>
+{
     /// Write the given `value` at the given `coord` of this [`SparseTensor`].
     async fn write_value(&self, txn_id: TxnId, coord: Coord, value: Number) -> TCResult<()>;
 }
@@ -179,7 +186,7 @@ where
         }
     }
 
-    async fn filled_at<'a>(self, txn: T, axes: Vec<usize>) -> TCResult<TCTryStream<'a, Coords>> {
+    async fn filled_at<'a>(self, txn: T, axes: Vec<usize>) -> TCResult<TCBoxTryStream<'a, Coords>> {
         match self {
             Self::Broadcast(broadcast) => broadcast.filled_at(txn, axes).await,
             Self::Cast(cast) => cast.filled_at(txn, axes).await,
@@ -259,19 +266,21 @@ where
             Self::Unary(unary) => unary.transpose(permutation).map(SparseAccess::accessor),
         }
     }
+}
 
+#[async_trait]
+impl<FD, FS, D, T> SparseWrite<FD, FS, D, T> for SparseAccessor<FD, FS, D, T>
+where
+    FD: File<Array> + TryFrom<D::File, Error = TCError>,
+    FS: File<Node> + TryFrom<D::File, Error = TCError>,
+    D: Dir,
+    T: Transaction<D>,
+    D::FileClass: From<TensorType>,
+{
     async fn write_value(&self, txn_id: TxnId, coord: Coord, value: Number) -> TCResult<()> {
         match self {
-            Self::Broadcast(broadcast) => broadcast.write_value(txn_id, coord, value).await,
-            Self::Cast(cast) => cast.write_value(txn_id, coord, value).await,
-            Self::Combine(combine) => combine.write_value(txn_id, coord, value).await,
-            Self::Dense(dense) => dense.write_value(txn_id, coord, value).await,
-            Self::Expand(expand) => expand.write_value(txn_id, coord, value).await,
-            Self::Reduce(reduce) => reduce.write_value(txn_id, coord, value).await,
-            Self::Slice(slice) => slice.write_value(txn_id, coord, value).await,
             Self::Table(table) => table.write_value(txn_id, coord, value).await,
-            Self::Transpose(transpose) => transpose.write_value(txn_id, coord, value).await,
-            Self::Unary(unary) => unary.write_value(txn_id, coord, value).await,
+            _ => Err(TCError::unsupported("cannot write to a Tensor view")),
         }
     }
 }
@@ -366,7 +375,11 @@ where
         Ok(Box::pin(filled))
     }
 
-    async fn filled_at<'a>(self, txn: T, axes: Vec<usize>) -> TCResult<TCTryStream<'a, Coords>> {
+    async fn filled_at<'a>(self, txn: T, axes: Vec<usize>) -> TCResult<TCBoxTryStream<'a, Coords>> {
+        if axes.is_empty() {
+            return Ok(Box::pin(stream::empty()));
+        }
+
         let bounds = Bounds::all(self.source.shape());
         let ndim = axes.len();
 
@@ -435,10 +448,6 @@ where
     fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self::Transpose> {
         self.source.transpose(permutation).map(DenseToSparse::from)
     }
-
-    async fn write_value(&self, txn_id: TxnId, coord: Coord, value: Number) -> TCResult<()> {
-        self.source.write_value(txn_id, coord.into(), value).await
-    }
 }
 
 impl<FD, FS, D, T, B> ReadValueAt<D> for DenseToSparse<FD, FS, D, T, B>
@@ -521,7 +530,7 @@ where
 impl<FD, FS, D, T, A> SparseAccess<FD, FS, D, T> for SparseBroadcast<FD, FS, D, T, A>
 where
     FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node>,
+    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
     A: SparseAccess<FD, FS, D, T>,
@@ -546,10 +555,11 @@ where
             return Ok(Box::pin(stream::empty()));
         }
 
-        let source_axes = (0..self.source.ndim()).collect();
+        let source_axes: Vec<usize> = (0..self.source.ndim()).collect();
         let ndim = self.ndim();
 
         let rebase = self.rebase.clone();
+
         let filled_at = self
             .source
             .clone()
@@ -559,7 +569,8 @@ where
         let coords = filled_at
             .map_ok(move |coords| stream::iter(coords.to_vec()).map(TCResult::Ok))
             .try_flatten()
-            .map_ok(move |coord| stream::iter(rebase.map_coord(coord).affected().map(TCResult::Ok)))
+            .map_ok(move |coord| rebase.map_coord(coord))
+            .map_ok(move |bounds| stream::iter(bounds.affected().map(TCResult::Ok)))
             .try_flatten();
 
         let coords = CoordBlocks::new(coords, ndim, PER_BLOCK);
@@ -567,10 +578,11 @@ where
         Ok(Box::pin(broadcast))
     }
 
-    async fn filled_at<'a>(self, txn: T, axes: Vec<usize>) -> TCResult<TCTryStream<'a, Coords>> {
-        debug!("SparseBroadcast::filled_at");
+    async fn filled_at<'a>(self, txn: T, axes: Vec<usize>) -> TCResult<TCBoxTryStream<'a, Coords>> {
+        debug!("SparseBroadcast::filled_at {:?}", axes);
+        self.shape().validate_axes(&axes)?;
 
-        if self.is_empty(&txn).await? {
+        if axes.is_empty() || self.is_empty(&txn).await? {
             return Ok(Box::pin(stream::empty()));
         }
 
@@ -579,19 +591,22 @@ where
             axes.iter().map(|x| shape[*x]).collect::<Vec<u64>>()
         });
 
+        let ndim = self.ndim();
         let rebase = self.rebase;
         let source_axes = (0..self.source.ndim()).collect();
-        let ndim = axes.len();
 
         let filled_at = self.source.filled_at(txn.clone(), source_axes).await?;
         let filled_at = filled_at
             .map_ok(|coords| stream::iter(coords.to_vec()).map(TCResult::Ok))
             .try_flatten()
-            .map_ok(move |coord| stream::iter(rebase.map_coord(coord).affected().map(TCResult::Ok)))
-            .try_flatten()
-            .map_ok(move |coord| axes.iter().map(|x| coord[*x]).collect());
+            .map_ok(move |coord| rebase.map_coord(coord))
+            // TODO: can this happen in `Coords`? maybe a new stream generator type?
+            .map_ok(move |bounds| stream::iter(bounds.affected().map(TCResult::Ok)))
+            .try_flatten();
+
+        let filled_at =
+            CoordBlocks::new(filled_at, ndim, PER_BLOCK).map_ok(move |coords| coords.get(&axes));
 
-        let filled_at = CoordBlocks::new(filled_at, ndim, PER_BLOCK);
         let filled_at = sorted_coords::<FD, FS, D, T, _>(&txn, shape, filled_at).await?;
 
         Ok(Box::pin(filled_at))
@@ -609,6 +624,7 @@ where
     }
 
     async fn is_empty(&self, txn: &T) -> TCResult<bool> {
+        debug!("SparseBroadcast::is_empty?");
         self.source.is_empty(txn).await
     }
 
@@ -624,10 +640,6 @@ where
     fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self::Transpose> {
         SparseTranspose::new(self, permutation)
     }
-
-    async fn write_value(&self, _txn_id: TxnId, _coord: Coord, _value: Number) -> TCResult<()> {
-        Err(TCError::unsupported(ERR_NONBIJECTIVE_WRITE))
-    }
 }
 
 impl<FD, FS, D, T, A> ReadValueAt<D> for SparseBroadcast<FD, FS, D, T, A>
@@ -642,13 +654,15 @@ where
     type Txn = T;
 
     fn read_value_at<'a>(self, txn: T, coord: Coord) -> Read<'a> {
-        let source_coord = self.rebase.invert_coord(&coord);
-        let read = self
-            .source
-            .read_value_at(txn, source_coord)
-            .map_ok(|(_, val)| (coord, val));
+        Box::pin(async move {
+            self.shape().validate_coord(&coord)?;
 
-        Box::pin(read)
+            let source_coord = self.rebase.invert_coord(&coord);
+            self.source
+                .read_value_at(txn, source_coord)
+                .map_ok(|(_, val)| (coord, val))
+                .await
+        })
     }
 }
 
@@ -721,7 +735,7 @@ where
         Ok(Box::pin(cast))
     }
 
-    async fn filled_at<'a>(self, txn: T, axes: Vec<usize>) -> TCResult<TCTryStream<'a, Coords>> {
+    async fn filled_at<'a>(self, txn: T, axes: Vec<usize>) -> TCResult<TCBoxTryStream<'a, Coords>> {
         self.source.filled_at(txn, axes).await
     }
 
@@ -749,10 +763,6 @@ where
             phantom: self.phantom,
         })
     }
-
-    async fn write_value(&self, txn_id: TxnId, coord: Coord, value: Number) -> TCResult<()> {
-        self.source.write_value(txn_id, coord, value).await
-    }
 }
 
 impl<FD, FS, D, T, A> ReadValueAt<D> for SparseCast<FD, FS, D, T, A>
@@ -898,8 +908,18 @@ where
         Ok(Box::pin(filled))
     }
 
-    async fn filled_at<'a>(self, txn: T, axes: Vec<usize>) -> TCResult<TCTryStream<'a, Coords>> {
-        let shape = self.shape().to_vec();
+    async fn filled_at<'a>(self, txn: T, axes: Vec<usize>) -> TCResult<TCBoxTryStream<'a, Coords>> {
+        self.shape().validate_axes(&axes)?;
+
+        if axes.is_empty() {
+            return Ok(Box::pin(stream::empty()));
+        }
+
+        let shape = {
+            let shape = self.shape();
+            axes.iter().map(|x| shape[*x]).collect()
+        };
+
         let (left, right) = try_join!(
             self.left.filled_at(txn.clone(), axes.clone()),
             self.right.filled_at(txn, axes)
@@ -923,6 +943,8 @@ where
     }
 
     fn slice(self, bounds: Bounds) -> TCResult<Self::Slice> {
+        debug!("SparseCombinator::slice {}", bounds);
+
         let left = self.left.slice(bounds.clone())?;
         let right = self.right.slice(bounds)?;
         assert_eq!(left.shape(), right.shape());
@@ -949,10 +971,6 @@ where
             phantom: self.phantom,
         })
     }
-
-    async fn write_value(&self, _txn_id: TxnId, _coord: Coord, _value: Number) -> TCResult<()> {
-        Err(TCError::unsupported(ERR_NONBIJECTIVE_WRITE))
-    }
 }
 
 impl<FD, FS, D, T, L, R> ReadValueAt<D> for SparseCombinator<FD, FS, D, T, L, R>
@@ -1054,12 +1072,32 @@ where
         Ok(Box::pin(filled))
     }
 
-    async fn filled_at<'a>(self, txn: T, axes: Vec<usize>) -> TCResult<TCTryStream<'a, Coords>> {
-        let axis = self.rebase.expand_axis();
+    async fn filled_at<'a>(self, txn: T, axes: Vec<usize>) -> TCResult<TCBoxTryStream<'a, Coords>> {
+        self.shape().validate_axes(&axes)?;
+
+        if axes.is_empty() {
+            return Ok(Box::pin(stream::empty()));
+        }
+
+        let mut i = 0;
+        let expand = loop {
+            if i >= axes.len() {
+                break None;
+            } else if axes[i] == self.rebase.expand_axis() {
+                break Some(i);
+            } else {
+                i += 1;
+            }
+        };
+
         let source_axes = self.rebase.invert_axes(axes);
         let source = self.source.filled_at(txn, source_axes).await?;
-        let filled_at = source.map_ok(move |coords| coords.expand_dim(axis));
-        Ok(Box::pin(filled_at))
+        if let Some(x) = expand {
+            let filled_at = source.map_ok(move |coords| coords.expand_dim(x));
+            Ok(Box::pin(filled_at))
+        } else {
+            Ok(source)
+        }
     }
 
     async fn filled_count(self, txn: T) -> TCResult<u64> {
@@ -1070,10 +1108,20 @@ where
         self.source.is_empty(txn).await
     }
 
-    fn slice(self, bounds: Bounds) -> TCResult<Self::Slice> {
+    fn slice(self, mut bounds: Bounds) -> TCResult<Self::Slice> {
+        debug!("SparseExpand {} slice {}", self.shape(), bounds);
         self.shape().validate_bounds(&bounds)?;
+        bounds.normalize(self.shape());
 
         let source_bounds = self.rebase.invert_bounds(bounds);
+
+        debug!(
+            "slice source {} with bounds {} (removed axis {})",
+            self.source.shape(),
+            source_bounds,
+            self.rebase.expand_axis()
+        );
+
         self.source.slice(source_bounds)
     }
 
@@ -1095,11 +1143,6 @@ where
         let source = self.source.transpose(permutation)?;
         SparseExpand::new(source, expand_axis)
     }
-
-    async fn write_value(&self, txn_id: TxnId, coord: Coord, value: Number) -> TCResult<()> {
-        let coord = self.rebase.invert_coord(&coord);
-        self.source.write_value(txn_id, coord, value).await
-    }
 }
 
 impl<FD, FS, D, T, A> ReadValueAt<D> for SparseExpand<FD, FS, D, T, A>
@@ -1211,6 +1254,7 @@ where
             .try_flatten()
             .map_ok(move |coord| {
                 let source_bounds = rebase.invert_coord(&coord);
+                debug!("reduce {:?} (bounds {})", coord, source_bounds);
                 let source = source.clone();
                 let txn = txn.clone();
                 Box::pin(async move {
@@ -1225,7 +1269,7 @@ where
         Ok(Box::pin(filled))
     }
 
-    async fn filled_at<'a>(self, txn: T, axes: Vec<usize>) -> TCResult<TCTryStream<'a, Coords>> {
+    async fn filled_at<'a>(self, txn: T, axes: Vec<usize>) -> TCResult<TCBoxTryStream<'a, Coords>> {
         let source_axes = self.rebase.invert_axes(axes);
         self.source.filled_at(txn, source_axes).await
     }
@@ -1245,10 +1289,13 @@ where
     }
 
     fn slice(self, bounds: Bounds) -> TCResult<Self::Slice> {
+        debug!("SparseReduce::slice {}", bounds);
+
         self.shape().validate_bounds(&bounds)?;
 
         let reduce_axis = self.rebase.reduce_axis(&bounds);
         let source_bounds = self.rebase.invert_bounds(bounds);
+        debug!("source bounds are {}", source_bounds);
         let source = self.source.slice(source_bounds)?;
         SparseReduce::new(source.into(), reduce_axis, self.reductor)
     }
@@ -1256,10 +1303,6 @@ where
     fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self::Transpose> {
         SparseTranspose::new(self, permutation)
     }
-
-    async fn write_value(&self, _txn_id: TxnId, _coord: Coord, _value: Number) -> TCResult<()> {
-        Err(TCError::unsupported(ERR_NONBIJECTIVE_WRITE))
-    }
 }
 
 impl<FD, FS, D, T> ReadValueAt<D> for SparseReduce<FD, FS, D, T>
@@ -1372,7 +1415,11 @@ where
         Ok(Box::pin(filled))
     }
 
-    async fn filled_at<'a>(self, txn: T, axes: Vec<usize>) -> TCResult<TCTryStream<'a, Coords>> {
+    async fn filled_at<'a>(self, txn: T, axes: Vec<usize>) -> TCResult<TCBoxTryStream<'a, Coords>> {
+        if axes.is_empty() {
+            return Ok(Box::pin(stream::empty()));
+        }
+
         let shape = self.shape().clone();
         let source_axes = self.rebase.invert_axes(axes);
         let permutation = self.rebase.map_axes(&source_axes);
@@ -1391,10 +1438,18 @@ where
     }
 
     fn slice(self, bounds: Bounds) -> TCResult<Self::Slice> {
+        debug!("SparseTranspose::slice {}", bounds);
         self.shape().validate_bounds(&bounds)?;
 
         let slice_permutation = self.rebase.invert_permutation(&bounds);
-        let source_bounds = self.rebase.invert_bounds(bounds);
+        debug!(
+            "slice permutation {}{} is {:?}",
+            self.shape(),
+            bounds,
+            slice_permutation
+        );
+
+        let source_bounds = self.rebase.invert_bounds(&bounds);
         let source = self.source.slice(source_bounds)?;
         source.transpose(Some(slice_permutation))
     }
@@ -1403,12 +1458,6 @@ where
         let permutation = permutation.map(|axes| self.rebase.invert_axes(axes));
         self.source.transpose(permutation)
     }
-
-    async fn write_value(&self, txn_id: TxnId, coord: Coord, value: Number) -> TCResult<()> {
-        self.source
-            .write_value(txn_id, self.rebase.invert_coord(&coord), value)
-            .await
-    }
 }
 
 impl<FD, FS, D, T, A> ReadValueAt<D> for SparseTranspose<FD, FS, D, T, A>
@@ -1501,7 +1550,7 @@ where
         Ok(Box::pin(cast))
     }
 
-    async fn filled_at<'a>(self, txn: T, axes: Vec<usize>) -> TCResult<TCTryStream<'a, Coords>> {
+    async fn filled_at<'a>(self, txn: T, axes: Vec<usize>) -> TCResult<TCBoxTryStream<'a, Coords>> {
         self.source.filled_at(txn, axes).await
     }
 
@@ -1530,10 +1579,6 @@ where
             transform: self.transform,
         })
     }
-
-    async fn write_value(&self, _txn_id: TxnId, _coord: Coord, _value: Number) -> TCResult<()> {
-        Err(TCError::unsupported(ERR_NONBIJECTIVE_WRITE))
-    }
 }
 
 impl<FD, FS, D, T> ReadValueAt<D> for SparseUnary<FD, FS, D, T>
